@@ -625,45 +625,19 @@ int nvme_set_features_async_event(int fd, __u32 events,
 int nvme_set_features_auto_pst(int fd, bool apste, bool save,
 			       struct nvme_feat_auto_pst *apst, __u32 *result)
 {
-	struct nvme_set_features_args args = {
-		.args_size = sizeof(args),
-		.fd = fd,
-		.fid = NVME_FEAT_FID_AUTO_PST,
-		.nsid = NVME_NSID_NONE,
-		.cdw11 = NVME_SET(!!apste, FEAT_APST_APSTE),
-		.save = save,
-		.uuidx = NVME_UUID_NONE,
-		.data = apst,
-		.data_len = sizeof(*apst),
-		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
-		.result = result,
-	};
-
-	return nvme_set_features(&args);
+	return nvme_set_features_data(fd, NVME_FEAT_FID_AUTO_PST,
+		NVME_NSID_NONE, NVME_SET(!!apste, FEAT_APST_APSTE), save,
+		sizeof(*apst), apst, result);
 }
 
 int nvme_set_features_timestamp(int fd, bool save, __u64 timestamp)
 {
 	__le64 t = cpu_to_le64(timestamp);
 	struct nvme_timestamp ts = {};
-	struct nvme_set_features_args args = {
-		.args_size = sizeof(args),
-		.fd = fd,
-		.fid = NVME_FEAT_FID_TIMESTAMP,
-		.nsid = NVME_NSID_NONE,
-		.cdw11 = 0,
-		.cdw12 = 0,
-		.save = save,
-		.uuidx = NVME_UUID_NONE,
-		.cdw15 = 0,
-		.data_len = sizeof(ts),
-		.data = &ts,
-		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
-		.result = NULL,
-	};
-
 	memcpy(ts.timestamp, &t, sizeof(ts.timestamp));
-	return nvme_set_features(&args);
+
+	return nvme_set_features_data(fd, NVME_FEAT_FID_TIMESTAMP,
+		NVME_NSID_NONE, 0, save, sizeof(ts), &ts, NULL);
 }
 
 int nvme_set_features_hctm(int fd, __u16 tmt2, __u16 tmt1,
@@ -764,23 +738,8 @@ int nvme_set_features_lba_sts_interval(int fd, __u16 lsiri, __u16 lsipi,
 int nvme_set_features_host_behavior(int fd, bool save,
 	struct nvme_feat_host_behavior *data)
 {
-	struct nvme_set_features_args args = {
-		.args_size = sizeof(args),
-		.fd = fd,
-		.fid = NVME_FEAT_FID_HOST_BEHAVIOR,
-		.nsid = NVME_NSID_NONE,
-		.cdw11 = 0,
-		.cdw12 = 0,
-		.save = false,
-		.uuidx = NVME_UUID_NONE,
-		.cdw15 = 0,
-		.data_len = sizeof(*data),
-		.data = data,
-		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
-		.result = NULL,
-	};
-
-	return nvme_set_features(&args);
+	return nvme_set_features_data(fd, NVME_FEAT_FID_HOST_BEHAVIOR,
+		NVME_NSID_NONE, 0, false, sizeof(*data), data, NULL);
 }
 
 int nvme_set_features_sanitize(int fd, bool nodrm, bool save, __u32 *result)
@@ -809,23 +768,9 @@ int nvme_set_features_host_id(int fd, bool exhid, bool save, __u8 *hostid)
 {
 	__u32 len = exhid ? 16 : 8;
 	__u32 value = !!exhid;
-	struct nvme_set_features_args args = {
-		.args_size = sizeof(args),
-		.fd = fd,
-		.fid = NVME_FEAT_FID_HOST_ID,
-		.nsid = NVME_NSID_NONE,
-		.cdw11 = value,
-		.cdw12 = 0,
-		.save = save,
-		.uuidx = NVME_UUID_NONE,
-		.cdw15 = 0,
-		.data_len = len,
-		.data = hostid,
-		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
-		.result = NULL,
-	};
 
-	return nvme_set_features(&args);
+	return nvme_set_features_data(fd, NVME_FEAT_FID_HOST_ID,
+		NVME_NSID_NONE, value, save, len, hostid, NULL);
 }
 
 int nvme_set_features_resv_mask(int fd, __u32 mask, bool save, __u32 *result)
@@ -1696,12 +1641,25 @@ int nvme_get_property(struct nvme_get_property_args *args)
 
 int nvme_sanitize_nvm(struct nvme_sanitize_nvm_args *args)
 {
-	__u32 cdw10 = NVME_SET(args->sanact, SANITIZE_CDW10_SANACT) |
-			NVME_SET(!!args->ause, SANITIZE_CDW10_AUSE) |
-			NVME_SET(args->owpass, SANITIZE_CDW10_OWPASS) |
-			NVME_SET(!!args->oipbp, SANITIZE_CDW10_OIPBP) |
-			NVME_SET(!!args->nodas, SANITIZE_CDW10_NODAS);
-	__u32 cdw11 = args->ovrpat;
+	const size_t size_v1 = sizeof_args(struct nvme_sanitize_nvm_args, nodas, __u64);
+	const size_t size_v2 = sizeof_args(struct nvme_sanitize_nvm_args, emvs, __u64);
+	__u32 cdw10, cdw11;
+
+	if (args->args_size < size_v1 || args->args_size > size_v2) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	cdw10 = NVME_SET(args->sanact, SANITIZE_CDW10_SANACT) |
+		NVME_SET(!!args->ause, SANITIZE_CDW10_AUSE) |
+		NVME_SET(args->owpass, SANITIZE_CDW10_OWPASS) |
+		NVME_SET(!!args->oipbp, SANITIZE_CDW10_OIPBP) |
+		NVME_SET(!!args->nodas, SANITIZE_CDW10_NODAS);
+
+	if (args->args_size == size_v2)
+		cdw10 |= NVME_SET(!!args->emvs, SANITIZE_CDW10_EMVS);
+
+	cdw11 = args->ovrpat;
 
 	struct nvme_passthru_cmd cmd = {
 		.opcode		= nvme_admin_sanitize_nvm,
@@ -1710,10 +1668,6 @@ int nvme_sanitize_nvm(struct nvme_sanitize_nvm_args *args)
 		.timeout_ms	= args->timeout,
 	};
 
-	if (args->args_size < sizeof(*args)) {
-		errno = EINVAL;
-		return -1;
-	}
 	return nvme_submit_admin_passthru(args->fd, &cmd, args->result);
 }
 
