@@ -41,6 +41,7 @@
  * @ENVME_CONNECT_CONNREFUSED:	connection refused
  * @ENVME_CONNECT_ADDRNOTAVAIL:	cannot assign requested address
  * @ENVME_CONNECT_IGNORED:	connect attempt is ignored due to configuration
+ * @ENVME_CONNECT_NOKEY:	the TLS key is missing
  */
 enum nvme_connect_err {
 	ENVME_CONNECT_RESOLVE	= 1000,
@@ -63,6 +64,7 @@ enum nvme_connect_err {
 	ENVME_CONNECT_CONNREFUSED,
 	ENVME_CONNECT_ADDRNOTAVAIL,
 	ENVME_CONNECT_IGNORED,
+	ENVME_CONNECT_NOKEY,
 };
 
 /**
@@ -483,7 +485,7 @@ static inline void nvme_feature_decode_namespace_write_protect(__u32 value,
 
 static inline void nvme_id_ns_flbas_to_lbaf_inuse(__u8 flbas, __u8 *lbaf_inuse)
 {
-	*lbaf_inuse = ((NVME_FLBAS_HIGHER(flbas) >> 1) |
+	*lbaf_inuse = ((NVME_FLBAS_HIGHER(flbas) << 4) |
 			NVME_FLBAS_LOWER(flbas));
 }
 
@@ -558,6 +560,8 @@ char *kv_keymatch(const char *kv, const char *key);
  */
 char *startswith(const char *s, const char *prefix);
 
+#define min(x, y) ((x) > (y) ? (y) : (x))
+
 #define __round_mask(val, mult) ((__typeof__(val))((mult)-1))
 
 /**
